# Frequencycounter-v01a.py(w) (09-12-2011) frequency counter
# For Python version 2.6 or 2.7
# With external module pyaudio (for Python version 2.6 or 2.7)
# Created by Onno Hoekstra (pa2ohh)
# Corrected for Linux by Tamas, HA5FTL
import pyaudio
import math
import time
import tkFont
from Tkinter import *
from tkFileDialog import askopenfilename
from tkFileDialog import asksaveasfilename
from tkSimpleDialog import askstring
from tkMessageBox import *


# Values that can be modified
CANVASwidth = 600           # Width of the canvas
CANVASheight = 200          # Height of the canvas

FREQprescaler = 4096.0      # Prescaling factor
FREQoffset = 0.0            # Offset frequency (kHz!)

SAMPLErate = 96000          # Sample rate of the soundcard 22050 44100 48000 88200 96000 192000
UPDATEspeed = 1.1           # Update speed, default 1.1, for slower PC's a higher value is perhaps required 

TIMEdiv1x = [0.01, 0.02, 0.05, 0.1, 0.2, 0.5, 1.0, 2.0, 5.0, 10.00] # Gate Time list in s/div
TIMEdiv = 4                 # index 4 from TIMEdiv1x as initial value


# Colors that can be modified
COLORframes = "#000080"     # Color = "#rrggbb" rr=red gg=green bb=blue, Hexadecimal values 00 - ff 
COLORcanvas = "#404040"
COLORtext = "#ffffff"
COLORaudiobar = "#606060"
COLORaudiook = "#00ff00"
COLORaudiomax = "#ff0000"
COLORzerocrossings = "#00ffff"

# Fontsizes that can be modified
FREQfontsize = 48           # Size of frequency text (Small=6, Large=48)
INFOfontsize = 8            # Size of info (Small=6, Large=24)

# Button sizes that can be modified
Buttonwidth1 = 12
Buttonwidth2 = 8

# Various global variables
AUDIOsignal1 = []           # Audio trace channel 1    
AUDIOdevin = None           # Audio device for input. None = Windows default
AUDIOdevout = None          # Audio device for output. None = Windows default

RUNstatus = 1               # 0 stopped, 1 start, 2 running, 3 stop and restart, 4 stop
AUDIOstatus = 0             # 0 audio off, 1 audio on

Ymin = 0                    # Minimum value of trace
Ymax = 0                    # Maximum value of trace
Z1 = 0                      # Value zerocrossing - 1
Z2 = 0                      # Value zerocrossing + 1

CH1frequency = 0.0          # Frequency channel 1
FREQunit = 1                # 0=Herz, 1=kiloHerz, 2=MegaHerz


# =================================== Start widgets routines ========================================
def Bnot():
    print "Routine not made yet"


def BFrequnit():
    global FREQunit
    global RUNstatus
    
    FREQunit = FREQunit + 1
    if FREQunit > 2:
        FREQunit = 0

    UpdateScreen()          # UpdateScreen() call 

 
def BAudiostatus():
    global AUDIOstatus
    global RUNstatus
    
    if (AUDIOstatus == 0):
        AUDIOstatus = 1
    else:
        AUDIOstatus = 0

    UpdateScreen()          # UpdateScreen() call 

   
def BStart():
    global RUNstatus
    
    if (RUNstatus == 0):
        RUNstatus = 1

    UpdateScreen()          # UpdateScreen() call 


def BStop():
    global RUNstatus
    
    if (RUNstatus == 1):
        RUNstatus = 0
    elif (RUNstatus == 2):
        RUNstatus = 3
    elif (RUNstatus == 4):
        RUNstatus = 3

    UpdateScreen()          # UpdateScreen() call 

   
def BSoundcard():
    global SAMPLErate
    global RUNstatus
    
    s = askstring("Sample rate","Value: " + str(SAMPLErate) + "\n\nNew value:\n(48000, 96000, 192000)")

    if (s == None):         # If Cancel pressed, then None
        return()

    try:                    # Error if for example no numeric characters or OK pressed without input (s = ""), then v = 0
        v = int(s)
    except:
        v = 0

    if v != 0:
        SAMPLErate = v

    if (RUNstatus == 2):    # Restart if running
        RUNstatus = 4

    UpdateScreen()          # UpdateScreen() call 

    
def BOffset():
    global FREQoffset
    
    s = askstring("Offset","Value: " + str(FREQoffset) + "\n\nNew value (kHz): ")

    if (s == None):         # If Cancel pressed, then None
        return()

    try:                    # Error if for example no numeric characters or OK pressed without input (s = "")
        v = float(s)
    except:
        s = "error"

    if s != "error":
        FREQoffset = v

    UpdateScreen()          # UpdateScreen() call

    
def BPrescaler():
    global FREQprescaler
    global CH1frequency
    
    s = askstring("Prescaler","Value: " + str(FREQprescaler) + "\n\nNew value: ")

    if (s == None):         # If Cancel pressed, then None and do the frequency calibration
        Cal = askyesno("Frequency calibration","Calibrate with measured frequency?", default = NO)
        if Cal == False:            # 
            return()
        else:
            s = askstring("Prescaler calibration","Measured audio frequency (Hz): " + str(CH1frequency) + "\n\nReal frequency (Hz): ")

            if (s == None):         # If Cancel pressed, then None
                return()

            try:                    # Error if for example no numeric characters or OK pressed without input (s = "")
                v = float(s)
            except:
                s = "error"

            if s != "error" and v != 0:
                FREQprescaler = v / CH1frequency

    else:                           # Manual input for the prescaler value
        try:                        # Error if for example no numeric characters or OK pressed without input (s = "")
            v = float(s)
        except:
            s = "error"

        if s != "error" and v != 0:
            FREQprescaler = v

    UpdateScreen()          # UpdateScreen() call
    
    
def BTime1():
    global TIMEdiv
    global RUNstatus
    
    if (TIMEdiv >= 1):
        TIMEdiv = TIMEdiv - 1

    if (RUNstatus == 2):    # Restart if running
        RUNstatus = 4

    UpdateScreen()          # UpdateScreen() call
    
    
def BTime2():
    global TIMEdiv1x
    global TIMEdiv
    global RUNstatus
    
    if (TIMEdiv < len(TIMEdiv1x) - 1):
        TIMEdiv = TIMEdiv + 1

    if (RUNstatus == 2):    # Restart if running
        RUNstatus = 4

    UpdateScreen()          # UpdateScreen() call
    
    
def BSave():
    global FREQprescaler
    global FREQoffset
    global SAMPLErate
    global TIMEdiv
    global FREQunit
    global RUNstatus
    
    filename = asksaveasfilename(filetypes=[("Setting","*.cfg"),("allfiles","*")])

    if (filename == None):              # No input, cancel pressed or an error
        filename = ""

    if (filename == ""):
        return()
    if filename[-4:] != ".cfg":
        filename = filename + ".cfg"
    Wfile = open(filename,'w')          # output file setting

    txt = "Frequency counter settings\n"
    Wfile.write(txt) 
    txt = str(FREQprescaler) + "\n"
    Wfile.write(txt)     
    txt = str(FREQoffset) + "\n"
    Wfile.write(txt)     
    txt = str(SAMPLErate) + "\n"
    Wfile.write(txt)     
    txt = str(TIMEdiv) + "\n"
    Wfile.write(txt)     
    txt = str(FREQunit) + "\n"
    Wfile.write(txt)     
    Wfile.close()           # Close the file

    if (RUNstatus == 2):    # Restart if running
        RUNstatus = 4

    UpdateScreen()          # UpdateScreen() call 


def BRecall():
    global FREQprescaler
    global FREQoffset
    global SAMPLErate
    global TIMEdiv
    global FREQunit
    global RUNstatus
    
    filename = askopenfilename(filetypes=[("Setting","*.cfg"),("allfiles","*")])

    if (filename == None):              # No input, cancel pressed or an error
        filename = ""

    if (filename == ""):
        return()

    Rfile = open(filename,'r')          # input file setting
    txt = Rfile.readline()              # read the first line, do nothing with it
    txt = Rfile.readline()              # read the next line
    FREQprescaler = float(txt)
    txt = Rfile.readline()              # read the next line
    FREQoffset = float(txt)
    txt = Rfile.readline()              # read the next line
    SAMPLErate = int(txt)
    txt = Rfile.readline()              # read the next line
    TIMEdiv = int(txt)
    txt = Rfile.readline()              # read the next line
    FREQunit = float(txt)
    txt = Rfile.readline()              # read the next line

    Rfile.close()                       # Close the file

    if (RUNstatus == 2):                # Restart if running
        RUNstatus = 4
       
    UpdateScreen()                      # UpdateScreen() call 
    
# ============================================ Main routine ====================================================
    
def AUDIOin():   # Read the audio from the stream and store the data into the arrays
    global AUDIOsignal1
    global AUDIOdevin
    global AUDIOdevout

    global RUNstatus
    global AUDIOstatus
    global TIMEdiv1x
    global TIMEdiv
    global SAMPLErate
    global SAMPLEratecalc
    global UPDATEspeed
    
    while (True):                                           # Main loop
        PA = pyaudio.PyAudio()
        FORMAT = pyaudio.paInt16                            # Audio format 16 levels and 2 channels
        CHUNK = int( float(SAMPLErate) * TIMEdiv1x[TIMEdiv])

        # RUNstatus = 1 : Open Stream
        if (RUNstatus == 1):
            if UPDATEspeed < 1:
                UPDATEspeed = 1.0
            if UPDATEspeed > 5:
                UPDATEspeed = 5.0

            TRACESopened = 1

            try:
                chunkbuffer = int(UPDATEspeed * CHUNK)

                if chunkbuffer < SAMPLErate / 10:           # Prevent buffer overload if small number of samples
                    chunkbuffer = int(SAMPLErate / 10)

                stream = PA.open(format = FORMAT,
                    channels = TRACESopened, 
                    rate = SAMPLErate, 
                    input = True,
                    output = True,
                    frames_per_buffer = int(chunkbuffer),
                    input_device_index = AUDIOdevin,
                    output_device_index = AUDIOdevout)
                
                RUNstatus = 2
            except:                                         # If error in opening audio stream, show error
                RUNstatus = 0
                txt = "Sample rate: " + str(SAMPLErate) + ", try a lower sample rate.\nOr another audio device."
                showerror("Cannot open Audio Stream", txt)

            UpdateScreen()                                  # UpdateScreen() call        

        
        # RUNstatus = 2: Reading audio data from soundcard
        if (RUNstatus == 2):
            signals = []
            try:
                signals = stream.read(chunkbuffer)          # Read samples from the buffer
            except:
                RUNstatus = 4
               
            if (AUDIOstatus == 1):                          # Audio on 
                stream.write(signals, chunkbuffer)

            # Start conversion audio samples to values -32762 to +32767 (one's complement)
            Lsignals = len(signals)                         # Lenght of signals array
            AUDIOsignal1 = []                               # Clear the AUDIOsignal1 array for trace 1

            Sbuffer = Lsignals / 2                          # Sbuffer is number of values (2 bytes per audio sample value, 1 channel is 2 bytes)
            i = 2 * int(Sbuffer - CHUNK)                    # Start value, first part is skipped due to possible distortions

            if i < 0:                                       # Prevent negative values for i
                i = 0
                
            s = Lsignals - 1       
            while (i < s):
                v = ord(signals[i]) + 256 * ord(signals[i+1])
                if v > 32767:                               # One's complement correction
                    v = v - 65535
                AUDIOsignal1.append(v)                      # Append the value to the trace 1 array 
                i = i + 2                                   # 2 bytes per sample value and 1 trace is 2 bytes totally

            UpdateAll()                                     # Update Data, trace and screen

        # RUNstatus = 3: Stop
        # RUNstatus = 4: Stop and restart
        if (RUNstatus == 3) or (RUNstatus == 4):
            stream.stop_stream()
            stream.close()
            PA.terminate()
            if RUNstatus == 3:
                RUNstatus = 0                               # Status is stopped 
            if RUNstatus == 4:          
                RUNstatus = 1                               # Status is (re)start

            UpdateScreen()                                  # Update screen with text  

        # Update tasks and screens by TKinter 
        root.update_idletasks()
        root.update()                                       # Activate updated screens 

  
def UpdateAll():        # Update Data, trace and screen
    MeasureFrequency()  # Frequency measurement
    UpdateScreen()      # Update screen with text


def UpdateScreen():     # Update screen with text
    MakeScreen()        # Update the text
    root.update()       # Activate updated screens    


def MeasureFrequency():             # Measure the frequency on channel 1
    global AUDIOsignal1
    global SAMPLErate
    global CH1frequency
    global Ymin
    global Ymax
    global Z1
    global Z2


    TRACEsize = len(AUDIOsignal1)   # Set the trace length
    if TRACEsize == 0:
        return()
    
    t = 0                           # Start finding minimum and maximum values from start of trace
    t1 = int(0.25 * TRACEsize)      # Stop finding minimum and maximum values at 0.25 x TRACEsize 

    Ymin = float(AUDIOsignal1[int(t)])
    Ymax = Ymin
    while (t < t1):
        v = float(AUDIOsignal1[int(t)])
        if v < Ymin:
            Ymin = v
        if v > Ymax:
            Ymax = v

        t = t + 1   

    # Frequency measurement routine
    if Ymax < 100 or Ymin > -100:   # No or low input signal
        CH1frequency = 0.0          # Set frequency to zero
        return()                    # and exit

    Tlevel = Ymax / 4               # Level for hysteresis at 1/4 of maximum value to avoid measuring noise around zero crossings
    t = 0                           # Start searching zero crossing from here
    t1 = int(0.25 * TRACEsize)      # First zero crossing between 0 and 0.25, otherwise skip measurement 

    while AUDIOsignal1[int(t)] < Tlevel and t < t1:     # Start search when signal is above Tlevel
        t = t + 1

    while AUDIOsignal1[int(t)] > 0 and t < t1:          # Find first sample below zero crossing, is Startsample
        t = t + 1
  
    Startsample = float(t)

    if Startsample >= t1:           # No zero crossing found
        CH1frequency = 0.0          # Set frequency to zero
        return()                    # and exit
    
    t2 = TRACEsize - 1              # Set the measurement period t2

    Nperiods = 0                    # Period counter
    Stopsample = Startsample        # Initialize Stopsample variable already to prevent errors if not done here below
    while t < t2:
        while AUDIOsignal1[int(t)] < Tlevel and t < t2:
            t = t + 1
        while AUDIOsignal1[int(t)] > 0 and t < t2:
            t = t + 1

        if t < t2:
            Stopsample = t
            Nperiods = Nperiods + 1

    # Make correction for fractions of samples, calculated from amplitude values or AUDIOsignal1
    u1a = float(AUDIOsignal1[int(Startsample-1)])
    u2a = float(-1 * AUDIOsignal1[int(Startsample)]) 
    u1b = float(AUDIOsignal1[int(Stopsample-1)])
    u2b = float(-1 * AUDIOsignal1[int(Stopsample)])

    Z1 = u1b                        # Value zerocrossing - 1
    Z2 = -1 * u2b                   # Value zerocrossing + 1

    Startsample = Startsample - u2a / (u1a + u2a)
    Stopsample = Stopsample - u2b / (u1b + u2b)

    if (Stopsample > Startsample):
        CH1frequency = float(SAMPLErate) * Nperiods /(Stopsample - Startsample)
    else:
        CH1frequency = 0.0
        

def MakeScreen():        # Update the text
    global AUDIOsignal1
    global RUNstatus
    global AUDIOstatus
    global TIMEdiv1x
    global TIMEdiv
    global SAMPLErate
    global CH1frequency
    global FREQunit
    global FREQprescaler
    global FREQoffset
    global FREQfont
    global INFOfont
    global Ymin
    global Ymax
    global Z1
    global Z2
    global CANVASwidth
    global CANVASheight


    # Delete all items on the screen
    de = ca.find_enclosed ( 0, 0, CANVASwidth+1000, CANVASheight+1000)   
    for n in de: 
        ca.delete(n)


    # General information on top of the screen
    if (AUDIOstatus == 1):
        txt = "Audio on "
        txt = txt + "    Max Level: " + str(int(Ymax / 320)) + "%    Min Level: " + str(int(Ymin/320)) + "%"
        txt = txt + "    Zerocrossing1: " + str(int(Z1/320)) + "%    Zerocrossing2: " + str(int(Z2/320)) + "%"
    else:
        txt = ""

    x = 10
    y = 20
    idTXT = ca.create_text (x, y, text=txt, font=INFOfont, anchor=W, fill=COLORtext)


    # Time gating information
    vx = TIMEdiv1x[TIMEdiv]
    if vx >= 1:
        txt = str(int(vx))
    elif vx >= 0.1:
        txt ="0." + str(int(vx * 10))
    elif vx >= 0.01:
        txt ="0.0" + str(int(vx * 100))

    txt = txt + " sec."
    if (RUNstatus == 0) or (RUNstatus == 3):
        txt = txt + "    stop"

    x = 10
    y = CANVASheight-20
    idTXT = ca.create_text (x, y, text=txt, font=INFOfont, anchor=W, fill=COLORtext)


    # Frequency counter information
    frequency = FREQprescaler * CH1frequency + 1000 * FREQoffset

    if frequency < 0:
        frequency = -1 * frequency
    
    if FREQunit == 0:
        frequency = frequency / 1
        txtunit = "Hz"
        FRround = 1        

    if FREQunit == 1:
        frequency = frequency / 1000
        txtunit = "kHz"
        FRround = 1

    if FREQunit == 2:
        frequency = frequency / 1000000
        txtunit = "MHz"
        FRround = 3    

    if FRround == 3:
        txt = str(int(frequency * 1000))
        while len(txt) < 4:
            txt = "0" + txt 
        txt = txt[:-3] + "." + txt[-3:]
        txt = txt + " " + txtunit

    if FRround == 2:
        txt = str(int(frequency * 100))
        while len(txt) < 3:
            txt = "0" + txt 
        txt = txt[:-2] + "." + txt[-2:]
        txt = txt + " " + txtunit

    if FRround == 1:
        txt = str(int(frequency * 10))
        while len(txt) < 2:
            txt = "0" + txt 
        txt = txt[:-1] + "." + txt[-1:]
        txt = txt + " " + txtunit

    if FRround == 0:
        txt = str(int(frequency))
        txt = txt + " " + txtunit

    x = 60
    y = CANVASheight / 2
    idTXT = ca.create_text(x, y, text=txt, font=FREQfont, anchor=W, fill=COLORtext)


def SELECTaudiodevice():        # Select an audio device
    global AUDIOdevin
    global AUDIOdevout

    PA = pyaudio.PyAudio()
    ndev = PA.get_device_count()

    n = 0
    ai = ""
    ao = ""
    while n < ndev:
        s = PA.get_device_info_by_index(n)
        # print n, s
        if s['maxInputChannels'] > 0:
            ai = ai + str(s['index']) + ": " + s['name'] + "\n"
        if s['maxOutputChannels'] > 0:
            ao = ao + str(s['index']) + ": " + s['name'] + "\n"
        n = n + 1
    PA.terminate()

    AUDIOdevin = None
    
    s = askstring("Device","Select audio INPUT device:\nPress Cancel for Windows Default\n\n" + ai + "\n\nNumber: ")
    if (s != None):         # If Cancel pressed, then None
        try:                    # Error if for example no numeric characters or OK pressed without input (s = "")
            v = int(s)
        except:
            s = "error"

        if s != "error":
            if v < 0 or v > ndev:
                v = 0
            AUDIOdevin = v

    AUDIOdevout = None

    s = askstring("Device","Select audio OUTPUT device:\nPress Cancel for Windows Default\n\n" + ao + "\n\nNumber: ")
    if (s != None):         # If Cancel pressed, then None
        try:                    # Error if for example no numeric characters or OK pressed without input (s = "")
            v = int(s)
        except:
            s = "error"

        if s != "error":
            if v < 0 or v > ndev:
                v = 0
            AUDIOdevout = v


# ================ Make Screen ==========================

root=Tk()
root.title("Frequency Counter V01a.py(w) (09-12-2011)")

root.minsize(100, 100)

frame1 = Frame(root, background=COLORframes, borderwidth=5, relief=RIDGE)
frame1.pack(side=TOP, expand=1, fill=X)

frame2 = Frame(root, background="black", borderwidth=5, relief=RIDGE)
frame2.pack(side=TOP, expand=1, fill=X)

frame3 = Frame(root, background=COLORframes, borderwidth=5, relief=RIDGE)
frame3.pack(side=TOP, expand=1, fill=X)

ca = Canvas(frame2, width=CANVASwidth, height=CANVASheight, background=COLORcanvas)
ca.pack(side=TOP)

b = Button(frame1, text="Hz-kHz-MHz", width=Buttonwidth1, command=BFrequnit)
b.pack(side=LEFT, padx=5, pady=5)

b = Button(frame1, text="Audio on/off", width=Buttonwidth1, command=BAudiostatus)
b.pack(side=LEFT, padx=5, pady=5)

b = Button(frame1, text="Soundcard", width=Buttonwidth1, command=BSoundcard)
b.pack(side=RIGHT, padx=5, pady=5)

b = Button(frame1, text="Offset", width=Buttonwidth1, command=BOffset)
b.pack(side=RIGHT, padx=5, pady=5)

b = Button(frame1, text="Prescaler", width=Buttonwidth1, command=BPrescaler)
b.pack(side=RIGHT, padx=5, pady=5)

b = Button(frame3, text="Start", width=Buttonwidth2, command=BStart)
b.pack(side=LEFT, padx=5, pady=5)

b = Button(frame3, text="Stop", width=Buttonwidth2, command=BStop)
b.pack(side=LEFT, padx=5, pady=5)

b = Button(frame3, text="-Time", width=Buttonwidth2, command=BTime1)
b.pack(side=LEFT, padx=5, pady=5)

b = Button(frame3, text="+Time", width=Buttonwidth2, command=BTime2)
b.pack(side=LEFT, padx=5, pady=5)

b = Button(frame3, text="Recall", width=Buttonwidth1, command=BRecall)
b.pack(side=RIGHT, padx=5, pady=5)

b = Button(frame3, text="Save", width=Buttonwidth1, command=BSave)
b.pack(side=RIGHT, padx=5, pady=5)

# Fonts, after initialisation of the tk screen! 
FREQfont = tkFont.Font(size=FREQfontsize)
# FREQfont = tkFont.Font(family="Helvetica", size=FREQfontsize, weight="bold")
INFOfont = tkFont.Font(size=INFOfontsize)
# INFOfont = tkFont.Font(family="Helvetica", size=INFOfontsize, weight="bold")

# ================ Call main routine ===============================
root.update()               # Activate updated screens
SELECTaudiodevice()
AUDIOin()
 


